import pytest
import numpy as np
import chainladder as cl
import pandas as pd

@pytest.fixture
def atol():
    return 1e-5


data = ["RAA", "ABC", "GenIns", "MW2008", "MW2014"]


def test_bk_fit_weight():
    """
    Test validation of sample_weight requirement. Should raise a value error if no weight is supplied.
    """
    raa = cl.load_sample("RAA")
    with pytest.raises(ValueError):
        cl.Benktander().fit(raa)

@pytest.mark.parametrize("data", data)
def test_benktander_to_chainladder(data, atol):
    tri = cl.load_sample(data)
    a = cl.Chainladder().fit(tri).ibnr_
    b = cl.Benktander(apriori=0.8, n_iters=255).fit(tri, sample_weight=a).ibnr_
    xp = tri.get_array_module()
    assert xp.allclose(xp.nan_to_num(a.values), xp.nan_to_num(b.values), atol=atol)


def test_bf_eq_cl_when_using_cl_apriori(qtr):
    cl_ult = cl.Chainladder().fit(qtr).ultimate_
    bf_ult = cl.BornhuetterFerguson().fit(qtr, sample_weight=cl_ult).ultimate_
    xp = cl_ult.get_array_module()
    assert xp.allclose(cl_ult.values, bf_ult.values, atol=1e-5)


def test_different_backends(clrd):
    clrd = clrd[["CumPaidLoss", "EarnedPremDIR"]]
    clrd = clrd[clrd["LOB"] == "wkcomp"]
    model = cl.BornhuetterFerguson().fit(
        clrd["CumPaidLoss"].sum().set_backend("numpy"),
        sample_weight=clrd["EarnedPremDIR"].sum().latest_diagonal.set_backend("numpy"),
    )
    assert (
        abs(
            (
                model.predict(
                    clrd["CumPaidLoss"].set_backend("sparse"),
                    sample_weight=clrd["EarnedPremDIR"].latest_diagonal.set_backend(
                        "sparse"
                    ),
                ).ibnr_.sum()
                - model.ibnr_
            ).sum()
        )
        < 1
    )

def test_odd_shaped_triangle():
    df = pd.DataFrame({
        "claim_year": 2000 + pd.Series([0] * 8 + [1] * 4),
        "claim_month": [1, 4, 7, 10] * 3,
        "dev_year": 2000 + pd.Series([0] * 4 + [1] * 8),
        "dev_month": [1, 4, 7, 10] * 3,
        "payment": [1] * 12,
    })
    tr = cl.Triangle(
        df,
        origin=["claim_year", "claim_month"],
        development=["dev_year", "dev_month"],
        columns="payment",
        cumulative=False,
    )
    atr = tr.grain("OYDQ")
    ult1 = cl.Benktander(apriori = 1,n_iters=10000).fit(cl.Development(average="volume").fit_transform(atr),sample_weight = atr.latest_diagonal).ultimate_.sum()
    ult2 = cl.Benktander(apriori = 1,n_iters=10000).fit(cl.Development(average="volume").fit_transform(tr),sample_weight = tr.latest_diagonal).ultimate_.grain("OYDQ").sum()
    assert abs(ult1 - ult2) < 1e-5


# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import copy
import dill
import json
import os
import numpy as np
import pandas as pd

from chainladder.utils.sparse import sp
from io import StringIO
from patsy import dmatrix # noqa
from sklearn.base import (
    BaseEstimator,
    TransformerMixin
)

from typing import (
    Iterable,
    Union,
    Optional,
    TYPE_CHECKING
)

if TYPE_CHECKING:
    from chainladder import Triangle
    from numpy.typing import ArrayLike
    from pandas import DataFrame
    from pandas.core.interchange.dataframe_protocol import DataFrame as DataFrameXchg
    from sparse import COO
    from types import ModuleType
    from typing import AnyStr
    from pandas._typing import (
        FilePath,
        ReadCsvBuffer
    )


def load_sample(
        key: str,
        *args,
        **kwargs
) -> Triangle:
    """Function to load datasets included in the chainladder package. These consist of CSV
    files located in the repository directory chainladder/utils/data.

    Parameters
    ----------
    key: str
        The name of the dataset, e.g. RAA, ABC, UKMotor, GenIns, etc. The name should match the
        file name, without extension, of one of the files in the sample data folder.

    Returns
    -------
        chainladder.Triangle of the loaded dataset.

    """
    from chainladder import Triangle


    # Set base path to be the parent directory of this file, e.g., the utils folder.
    utils_path: AnyStr = os.path.dirname(os.path.abspath(__file__))

    # Validate that the file indicated by the key argument exists.
    dataset_path: str = os.path.join(utils_path, "data", key.lower() + ".csv")

    if not os.path.exists(dataset_path):
        raise ValueError(
            """
            Invalid key supplied. The key should match the name, without extension, of one of the file names
            in the sample data set folder. Please refer to the documentation page on sample data sets to see 
            what data are available.
            """
         )

    # Set initial values for arguments to Triangle __init__. These may be overridden by
    # values specific to the data set.
    origin: str = "origin"
    development: str = "development"
    columns: list = ["values"]
    index: list | None = None
    cumulative: bool = True

    if key.lower() in [
        "mcl",
        "usaa",
        "quarterly",
        "auto",
        "usauto",
        "tail_sample"
    ]:
        columns: list = [
            "incurred",
            "paid"
        ]
    if key.lower() == "clrd":
        origin: str = "AccidentYear"
        development: str = "DevelopmentYear"
        index: list = [
            "GRNAME",
            "LOB"
        ]
        columns: list = [
            "IncurLoss",
            "CumPaidLoss",
            "BulkLoss",
            "EarnedPremDIR",
            "EarnedPremCeded",
            "EarnedPremNet",
        ]
    if key.lower() == "berqsherm":
        origin: str = "AccidentYear"
        development: str = "DevelopmentYear"
        index: list = ["LOB"]
        columns: list = [
            "Incurred",
            "Paid",
            "Reported",
            "Closed"
        ]
    if key.lower() == "xyz":
        origin: str = "AccidentYear"
        development: str = "DevelopmentYear"
        columns: list = [
            "Incurred",
            "Paid",
            "Reported",
            "Closed",
            "Premium"
        ]
    if key.lower() in [
        "liab",
        "auto"
    ]:
        index: list = ["lob"]
    if key.lower() in [
        "cc_sample",
        "ia_sample"
    ]:
        columns: list = [
            "loss",
            "exposure"
        ]
    if key.lower() in ["prism"]:
        columns: list = [
            "reportedCount",
            "closedPaidCount",
            "Paid",
            "Incurred"
        ]
        index: list = [
            "ClaimNo",
            "Line",
            "Type",
            "ClaimLiability",
            "Limit",
            "Deductible"
        ]
        origin: str = "AccidentDate"
        development: str = "PaymentDate"
        cumulative: bool = False

    df = pd.read_csv(filepath_or_buffer=dataset_path)

    return Triangle(
        data=df,
        origin=origin,
        development=development,
        index=index,
        columns=columns,
        cumulative=cumulative,
        *args,
        **kwargs
    )


def read_pickle(path):
    with open(path, "rb") as pkl:
        return dill.load(pkl)

def read_csv(
        filepath_or_buffer: FilePath | ReadCsvBuffer[bytes] | ReadCsvBuffer[str],
        origin: Optional[str | list] = None,
        development: Optional[str | list] = None,
        columns: Optional[str | list] = None,
        index: Optional[str | list] = None,
        origin_format: Optional[str] = None,
        development_format: Optional[str] = None,
        cumulative: Optional[bool] = None,
        array_backend: str = None,
        pattern=False,
        trailing: bool = True,
        *args, 
        **kwargs
        ) -> Triangle:
    """
    Funtion that creates Triangle directly from input. Wrapper for pandas dataframe:
    https://pandas.pydata.org/docs/reference/api/pandas.read_csv.html

    Parameters
    ----------
    filepath_or_buffer: str, path object or file-like object
        Any valid string path is acceptable. The string could be a URL. Valid URL schemes 
        include http, ftp, s3, gs, and file. For file URLs, a host is expected. A local 
        file could be: file://localhost/path/to/table.csv.
        If you want to pass in a path object, pandas accepts any os.PathLike.
        By file-like object, we refer to objects with a read() method, such as a 
        file handle (e.g. via builtin open function) or StringIO.
    origin: str or list
         A representation of the accident, reporting or more generally the
         origin period of the triangle that will map to the Origin dimension
    development: str or list
        A representation of the development/valuation periods of the triangle
        that will map to the Development dimension
    columns: str or list
        A representation of the numeric data of the triangle that will map to
        the columns dimension.  If None, then a single 'Total' key will be
        generated.
    index: str or list or None
        A representation of the index of the triangle that will map to the
        index dimension.  If None, then a single 'Total' key will be generated.
    origin_format: optional str
        A string representation of the date format of the origin arg. If
        omitted then date format will be inferred by pandas.
    development_format: optional str
        A string representation of the date format of the development arg. If
        omitted then date format will be inferred by pandas.
    cumulative: bool
        Whether the triangle is cumulative or incremental.  This attribute is
        required to use the ``grain`` and ``dev_to_val`` methods and will be
        automatically set when invoking ``cum_to_incr`` or ``incr_to_cum`` methods.
    trailing: bool
        When partial origin periods are present, setting trailing to True will
        ensure the most recent origin period is a full period and the oldest
        origin is partial. If full origin periods are present in the data, then
        trailing has no effect.

    Attributes
    ----------
    index: Series
        Represents all available levels of the index dimension.
    columns: Series
        Represents all available levels of the value dimension.
    origin: DatetimeIndex
        Represents all available levels of the origin dimension.
    development: Series
        Represents all available levels of the development dimension.
    key_labels: list
        Represents the ``index`` axis labels
    virtual_columns: Series
        Represents the subset of columns of the triangle that are virtual.
    valuation: DatetimeIndex
        Represents all valuation dates of each cell in the Triangle.
    origin_grain: str
        The grain of the origin vector ('Y', 'S', 'Q', 'M')
    development_grain: str
        The grain of the development vector ('Y', 'S', 'Q', 'M')
    shape: tuple
        The 4D shape of the triangle instance with axes corresponding to (index, columns, origin, development)
    link_ratio, age_to_age
        Displays age-to-age ratios for the triangle.
    valuation_date : date
        The latest valuation date of the data
    loc: Triangle
        pandas-style ``loc`` accessor
    iloc: Triangle
        pandas-style ``iloc`` accessor
    latest_diagonal: Triangle
        The latest diagonal of the triangle
    is_cumulative: bool
        Whether the triangle is cumulative or not
    is_ultimate: bool
        Whether the triangle has an ultimate valuation
    is_full: bool
        Whether lower half of Triangle has been filled in
    is_val_tri:
        Whether the triangle development period is expressed as valuation
        periods.
    values: array
        4D numpy array underlying the Triangle instance
    T: Triangle
        Transpose index and columns of object.  Only available when Triangle is
        convertible to DataFrame.
    """


    from chainladder import Triangle

    #Chainladder implementation of: https://pandas.pydata.org/docs/reference/api/pandas.read_csv.html
    #This will allow the user to create a trignel directly from csv instead in csv -> dataframe -> triangle

    #create a data frame using the *args and **kwargs that the user specified
    local_dataframe = pd.read_csv(filepath_or_buffer,*args, **kwargs)

    #pass the created local_dataframe in the Triangle constructor 
    local_triangle = Triangle(
        data = local_dataframe, 
        origin=origin,
        development=development,
        columns=columns,
        index=index,
        origin_format=origin_format,
        development_format=development_format,
        cumulative=cumulative,
        array_backend=array_backend,
        pattern=pattern,
        trailing = trailing
    )

    return local_triangle

def read_json(json_str, array_backend=None):
    from chainladder import Triangle
    from chainladder.workflow import Pipeline

    if array_backend is None:
        from chainladder import options

        array_backend = options.ARRAY_BACKEND
    json_dict = json.loads(json_str)
    if type(json_dict) is list:
        import chainladder as cl

        return Pipeline(
            steps=[
                (
                    item["name"],
                    cl.__dict__[item["__class__"]]().set_params(**item["params"]),
                )
                for item in json_dict
            ]
        )
    elif "metadata" in json_dict.keys():
        j = json.loads(json_str)
        y = pd.read_json(StringIO(j["data"]), orient="split", date_unit="ns")
        y["origin"] = pd.to_datetime(y["origin"])
        y.columns = [c if c != "valuation" else "development" for c in y.columns]
        y["development"] = pd.to_datetime(y["development"])
        index = list(y.columns[: list(y.columns).index("origin")])
        columns = list(y.columns[list(y.columns).index("development") + 1 :])
        tri = Triangle(
            y,
            origin="origin",
            development="development",
            index=index,
            columns=columns,
            pattern=json.loads(j["metadata"])["is_pattern"],
            cumulative=False,
        )
        if json.loads(j["metadata"])["is_val_tri"]:
            tri = tri.dev_to_val()
        if json.loads(j["metadata"])["is_cumulative"]:
            tri = tri.incr_to_cum()
        tri = tri[json.loads(j["metadata"])["columns"]].sort_index()
        if "sub_tris" in json_dict.keys():
            for k, v in json_dict["sub_tris"].items():
                setattr(tri, k, read_json(v, array_backend))
                setattr(getattr(tri, k), "origin_grain", tri.origin_grain)
                setattr(getattr(tri, k), "development_grain", tri.development_grain)
        if "dfs" in json_dict.keys():
            for k, v in json_dict["dfs"].items():
                df = pd.read_json(v)
                if len(df.columns) == 1:
                    df = df.iloc[:, 0]
                setattr(tri, k, df)
        if array_backend:
            return tri.set_backend(array_backend)
        else:
            return tri
    else:
        import chainladder as cl

        return cl.__dict__[json_dict["__class__"]]().set_params(**json_dict["params"])


def parallelogram_olf(
    values,
    date,
    start_date=None,
    end_date=None,
    grain="Y",
    approximation_grain="M",
    vertical_line=False,
):
    """Parallelogram approach to on-leveling."""
    date = pd.to_datetime(date)
    if not start_date:
        start_date = "{}-01-01".format(date.min().year)
    if not end_date:
        end_date = "{}-12-31".format(date.max().year)
    start_date = pd.to_datetime(start_date) - pd.tseries.offsets.DateOffset(days=1)

    date_freq = {
        "M": "MS",
        "D": "D",
    }
    if approximation_grain not in ['M', 'D']:
        raise ValueError("approximation_grain must be " "M" " or " "D" "")
    date_idx = pd.date_range(
        start_date - pd.tseries.offsets.DateOffset(years=1),
        end_date,
        freq=date_freq[approximation_grain],
    )

    rate_changes = pd.Series(np.array(values), np.array(date))
    rate_changes = rate_changes.reindex(date_idx, fill_value=0)

    cum_rate_changes = np.cumprod(1 + rate_changes.values)
    cum_rate_changes = pd.Series(cum_rate_changes, rate_changes.index)
    crl = cum_rate_changes.iloc[-1]

    cum_avg_rate_non_leaps = cum_rate_changes
    cum_avg_rate_leaps = cum_rate_changes

    if not vertical_line:
        rolling_num = {
            "M": 12,
            "D": 365,
        }

        cum_avg_rate_non_leaps = cum_rate_changes.rolling(
            rolling_num[approximation_grain]
        ).mean()
        cum_avg_rate_non_leaps = (
            cum_avg_rate_non_leaps + cum_avg_rate_non_leaps.shift(1).values
        ) / 2

        cum_avg_rate_leaps = cum_rate_changes.rolling(
            rolling_num[approximation_grain] + 1
        ).mean()
        cum_avg_rate_leaps = (
            cum_avg_rate_leaps + cum_avg_rate_leaps.shift(1).values
        ) / 2

    dropdates_num = {
        "M": 12,
        "D": 366,
    }
    cum_avg_rate_non_leaps = cum_avg_rate_non_leaps.iloc[
        dropdates_num[approximation_grain] :
    ]
    cum_avg_rate_leaps = cum_avg_rate_leaps.iloc[
        dropdates_num[approximation_grain] + 1 :
    ]

    fcrl_non_leaps = (
        cum_avg_rate_non_leaps.groupby(cum_avg_rate_non_leaps.index.to_period(grain))
        .mean()
        .reset_index()
    )
    fcrl_non_leaps.columns = ["Origin", "OLF"]
    fcrl_non_leaps["Origin"] = fcrl_non_leaps["Origin"].astype(str)
    fcrl_non_leaps["OLF"] = crl / fcrl_non_leaps["OLF"]

    fcrl_leaps = (
        cum_avg_rate_leaps.groupby(cum_avg_rate_leaps.index.to_period(grain))
        .mean()
        .reset_index()
    )
    fcrl_leaps.columns = ["Origin", "OLF"]
    fcrl_leaps["Origin"] = fcrl_leaps["Origin"].astype(str)
    fcrl_leaps["OLF"] = crl / fcrl_leaps["OLF"]

    combined = fcrl_non_leaps.join(fcrl_leaps, lsuffix="_non_leaps", rsuffix="_leaps")
    combined["is_leap"] = pd.to_datetime(
        combined["Origin_non_leaps"], format="%Y" + ("-%M" if grain == "M" else "")
    ).dt.is_leap_year
    

    if approximation_grain == "M":
        combined["final_OLF"] = combined["OLF_non_leaps"]
    else:
        combined["final_OLF"] = np.where(
            combined["is_leap"], combined["OLF_leaps"], combined["OLF_non_leaps"]
        )

    combined.drop(
        ["OLF_non_leaps", "Origin_leaps", "OLF_leaps", "is_leap"],
        axis=1,
        inplace=True,
    )
    combined.columns = ["Origin", "OLF"]

    return combined.set_index("Origin")


def set_common_backend(objs):
    from chainladder import options

    priority = options.ARRAY_PRIORITY
    backend = priority[np.min([priority.index(i.array_backend) for i in objs])]
    return [i.set_backend(backend) for i in objs]


def concat(
    objs: Iterable,
    axis: Union[int, str],
    ignore_index: bool = False,
    sort: bool = False,
):
    """Concatenate Triangle objects along a particular axis.

    Parameters
    ----------
    objs: list or tuple
        A list or tuple of Triangle objects to concat. All non-concat axes must
        be identical and all elements of the concat axes must be unique.
    axis: string or int
        The axis to concatenate along.
    ignore_index: bool, default False
        If True, do not use the index values along the concatenation axis. The
        resulting axis will be labeled 0, …, n - 1. This is useful if you are
        concatenating objects where the concatenation axis does not have
        meaningful indexing information. Note the index values on the other
        axes are still respected in the join.

    Returns
    -------
    Updated triangle
    """
    if type(objs) not in (list, tuple):
        raise TypeError("objects to be concatenated must be in a list or tuple")
    if type(objs) is tuple:
        objs = list(objs)
    if len(objs) == 0:
        raise ValueError("objs must contain at least one element")
    xp = objs[0].get_array_module()
    axis = objs[0]._get_axis(axis)

    if axis != 1:
        all_columns = []
        for obj in objs:
            for col in obj.columns:
                if col not in all_columns:
                    all_columns.append(col)
        for num, obj in enumerate(objs):
            for col in all_columns:
                if col not in obj.columns:
                    obj = copy.deepcopy(obj)
                    obj[col] = xp.nan
                    objs[num] = obj
    objs = set_common_backend(objs)
    mapper = {0: "kdims", 1: "vdims", 2: "odims", 3: "ddims"}
    for k in mapper.keys():
        if k != axis and k != 1:  # All non-concat axes must be identical
            a = np.array([getattr(obj, mapper[k]) for obj in objs])
            assert np.all(a == a[0])
        else:  # All elements of concat axis must be unique
            if ignore_index:
                new_axis = np.arange(
                    np.sum([len(getattr(obj, mapper[axis])) for obj in objs])
                )
                new_axis = new_axis[:, None] if axis == 0 else new_axis
            else:
                new_axis = np.concatenate([getattr(obj, mapper[axis]) for obj in objs])
            if axis == 0:
                assert len(pd.DataFrame(new_axis).drop_duplicates()) == len(new_axis)
            else:
                assert len(new_axis) == len(set(new_axis))
    out = copy.deepcopy(objs[0])
    out.values = xp.concatenate([obj.values for obj in objs], axis=axis)
    setattr(out, mapper[axis], new_axis)
    if ignore_index and axis == 0:
        out.key_labels = ["Index"]
    out.valuation_date = pd.Series([obj.valuation_date for obj in objs]).max()
    if out.ddims.dtype == "datetime64[ns]" and type(out.ddims) == np.ndarray:
        out.ddims = pd.DatetimeIndex(out.ddims)
    out._set_slicers()
    if sort:
        return out.sort_axis(axis)
    else:
        return out


def num_to_value(
        arr: ArrayLike,
        value
) -> ArrayLike:
    """
    Function that turns all zeros to nan values in an array.
    """
    backend = arr.__class__.__module__.split(".")[0]
    if backend == "sparse":
        if arr.fill_value == 0 or sp.isnan(arr.fill_value):
            arr.coords = arr.coords[:, arr.data != 0]
            arr.data = arr.data[arr.data != 0]

            arr: COO = sp(
                coords=arr.coords,
                data=arr.data,
                fill_value=sp.nan, # noqa
                shape=arr.shape
            )
        else:
            arr: COO = sp(
                num_to_nan(np.nan_to_num(arr.todense())),
                fill_value=value
            )
    else:
        arr[arr == 0] = value
    return arr


def num_to_nan(arr: ArrayLike) -> ArrayLike:
    """
    Function that turns all zeros to nan values in an array.

    Parameters
    ----------
    arr: ArrayLike
        An array-like object. For example, the values used in a Triangle.

    Returns
    -------
        The supplied array with all zeros converted to nan values. These nans
        are specific to the Triangle backend used.
    """

    from chainladder import Triangle

    # Take the nan specific to the module of the backend used. e.g., numpy, cupy, sparse, etc.
    xp: ModuleType = Triangle.get_array_module(
        None,
        arr=arr
    )

    return num_to_value(arr, xp.nan)


def minimum(x1, x2):
    return x1.minimum(x2)


def maximum(x1, x2):
    return x1.maximum(x2)

def to_period(dateseries: pd.Series, freq:str):
    if freq[:2] != '2Q':
        return dateseries.dt.to_period(freq)
    else:
        return dateseries.where(dateseries.dt.to_period(freq).dt.strftime('%q').isin(['1','3']),dateseries.dt.date + pd.DateOffset(months=-3)).dt.to_period(freq)

class PatsyFormula(BaseEstimator, TransformerMixin):
    """A sklearn-style Transformer for patsy formulas.

    PatsyFormula allows for R-style formula preprocessing of the ``design_matrix``
    of a machine learning algorithm. It's particularly useful with the `DevelopmentML`
    and `TweedieGLM` estimators.

    Parameters
    ----------

    formula: str
        A string representation of the regression model X features.

    Attributes
    ----------
    design_info_:
        The patsy instructions for generating the design_matrix, X.

    """

    def __init__(self, formula=None):
        self.formula = formula

    def _check_X(self, X):
        from chainladder.core import Triangle

        if isinstance(X, Triangle):
            raise AttributeError("X must be a pandas dataframe, not a Triangle")

    def fit(self, X, y=None, sample_weight=None):
        self._check_X(X)
        self.design_info_ = dmatrix(self.formula, X).design_info
        return self

    def transform(self, X):
        self._check_X(X)
        return dmatrix(self.design_info_, X)


def model_diagnostics(model, name=None, groupby=None):
    """A helper function that summarizes various vectors of an
    IBNR model as columns of a Triangle

    Parameters
    ----------
    model:
        A chainladder IBNR estimator or Pipeline
    name:
        An alias to give the model. This will be added to the index of the return
        triangle.
    groupby:
        The index level at which the model should be summarized

    Returns
    -------
    Triangle up select origin vectors, IBNR, ultimate, Latest diagonal, etc.
    """
    from chainladder import Pipeline

    if isinstance(model, Pipeline):
        obj = copy.deepcopy(model.steps[-1][-1])
    else:
        obj = copy.deepcopy(model)
    if groupby is not None:
        obj.X_ = obj.X_.groupby(groupby).sum().cum_to_incr()
        obj.ultimate_ = obj.ultimate_.groupby(groupby).sum()
        if hasattr(obj, "expectation_"):
            obj.expectation_ = obj.expectation_.groupby(groupby).sum()
    else:
        obj.X_ = obj.X_.incr_to_cum()
    val = obj.X_.valuation
    latest = obj.X_.sum("development")
    run_off = obj.full_expectation_.iloc[..., :-1].dev_to_val().cum_to_incr()
    run_off = run_off[run_off.development > str(obj.X_.valuation_date)]
    run_off = run_off.iloc[
        ..., : {"M": 12, "S": 6, "Q": 4, "Y": 1}[obj.X_.development_grain]
    ]

    triangles = []
    for col in obj.ultimate_.columns:
        idx = latest.index
        idx["Measure"] = col
        idx["Model"] = obj.__class__.__name__ if name is None else name
        idx = idx[list(idx.columns[-2:]) + list(idx.columns[:-2])]
        out = latest[col].rename("columns", ["Latest"])
        if obj.X_.development_grain in ["M"]:
            out["Month Incremental"] = obj.X_[col][val == obj.X_.valuation_date].sum(
                "development"
            )
        if (
            obj.X_.development_grain in ["M", "Q"]
            and pd.Period(out.valuation_date, freq="Q").to_timestamp(how="s")
            > val.min()
        ):
            out["Quarter Incremental"] = (
                obj.X_
                - obj.X_[
                    val
                    < pd.Period(out.valuation_date, freq="Q")
                    .to_timestamp(how="s")
                    .strftime("%Y-%m")
                ]
            ).sum("development")[col]
        else:
            out["Quarter Incremental"] = 0
        if pd.Period(out.valuation_date, freq="Y").to_timestamp(how="s") > val.min():
            out["Year Incremental"] = (
                obj.X_ - obj.X_[val < str(obj.X_.valuation_date.year)]
            ).sum("development")[col]
        else:
            out["Year Incremental"] = 0
        out["IBNR"] = obj.ibnr_[col]
        out["Ultimate"] = obj.ultimate_[col]
        for i in range(run_off.shape[-1]):
            out["Run Off " + str(i + 1)] = run_off[col].iloc[..., i]
        if hasattr(obj, "expectation_"):
            out["Apriori"] = (
                obj.expectation_
                if obj.expectation_.shape[1] == 1
                else obj.expectation_[col]
            )
        out.index = idx
        triangles.append(out)
    return concat(triangles, 0)



### Building out a dev environment with a working copy
### of R ChainLadder is difficult.  These tests are 
### Currently inactive, but available should the compatibility
### of the installs improve at a later date.

import numpy as np
import chainladder as cl
import pytest

try:
    from rpy2.robjects.packages import importr
    from rpy2.robjects import r
    CL = importr("ChainLadder")
except:
    pass


@pytest.mark.r
def test_mcl_paid():
    df = r("MunichChainLadder(MCLpaid, MCLincurred)").rx("MCLPaid")
    p = cl.MunichAdjustment(paid_to_incurred=("paid", "incurred")).fit(
        cl.Development(sigma_interpolation="mack").fit_transform(cl.load_sample("mcl"))
    )
    xp = p.ldf_.get_array_module()
    arr = xp.array(df[0])
    assert xp.allclose(arr, p.munich_full_triangle_[0, 0, 0, :, :], atol=1e-5)


@pytest.mark.r
def test_mcl_incurred():
    df = r("MunichChainLadder(MCLpaid, MCLincurred)").rx("MCLIncurred")
    p = cl.MunichAdjustment(paid_to_incurred=[("paid", "incurred")]).fit(
        cl.Development(sigma_interpolation="mack").fit_transform(cl.load_sample("mcl"))
    )
    xp = p.ldf_.get_array_module()
    arr = xp.array(df[0])
    assert xp.allclose(arr, p.munich_full_triangle_[1, 0, 0, :, :], atol=1e-5)


### Building out a dev environment with a working copy
### of R ChainLadder is difficult.  These tests are 
### Currently inactive, but available should the compatibility
### of the installs improve at a later date.

import numpy as np
import chainladder as cl
import pytest

try:
    from rpy2.robjects.packages import importr
    from rpy2.robjects import r
    CL = importr("ChainLadder")
except:
    pass

@pytest.mark.r
def test_clarkldf(genins):
    model = cl.ClarkLDF().fit(genins)
    df = r("ClarkLDF(GenIns)").rx("THETAG")
    r_omega = df[0][0]
    r_theta = df[0][1]
    assert abs(model.omega_.iloc[0, 0] - r_omega) < 1e-2
    assert abs(model.theta_.iloc[0, 0] / 12 - r_theta) < 1e-2

@pytest.mark.r
def test_clarkldf_weibull(genins):
    model = cl.ClarkLDF(growth="weibull").fit(genins)
    df = r('ClarkLDF(GenIns, G="weibull")').rx("THETAG")
    r_omega = df[0][0]
    r_theta = df[0][1]
    assert abs(model.omega_.iloc[0, 0] - r_omega) < 1e-2
    assert abs(model.theta_.iloc[0, 0] / 12 - r_theta) < 1e-2

@pytest.mark.r
def test_clarkcapecod(genins):
    df = r("ClarkCapeCod(GenIns, Premium=10000000+400000*0:9)")
    r_omega = df.rx("THETAG")[0][0]
    r_theta = df.rx("THETAG")[0][1]
    r_elr = df.rx("ELR")[0][0]
    premium = genins.latest_diagonal * 0 + 1
    premium.values = (np.arange(10) * 400000 + 10000000)[None, None, :, None]
    model = cl.ClarkLDF().fit(genins, sample_weight=premium)
    assert abs(model.omega_.iloc[0, 0] - r_omega) < 1e-2
    assert abs(model.theta_.iloc[0, 0] / 12 - r_theta) < 1e-2
    assert abs(model.elr_.iloc[0, 0] - r_elr) < 1e-2

@pytest.mark.r
def test_clarkcapcod_weibull(genins):
    df = r('ClarkCapeCod(GenIns, Premium=10000000+400000*0:9, G="weibull")')
    r_omega = df.rx("THETAG")[0][0]
    r_theta = df.rx("THETAG")[0][1]
    r_elr = df.rx("ELR")[0][0]
    premium = genins.latest_diagonal * 0 + 1
    premium.values = (np.arange(10) * 400000 + 10000000)[None, None, :, None]
    model = cl.ClarkLDF(growth="weibull").fit(genins, sample_weight=premium)
    assert abs(model.omega_.iloc[0, 0] - r_omega) < 1e-2
    assert abs(model.theta_.iloc[0, 0] / 12 - r_theta) < 1e-2
    assert abs(model.elr_.iloc[0, 0] - r_elr) < 1e-2



### Building out a dev environment with a working copy
### of R ChainLadder is difficult.  These tests are 
### Currently inactive, but available should the compatibility
### of the installs improve at a later date.

import numpy as np
import pytest
import chainladder as cl

try:
    from rpy2.robjects.packages import importr
    from rpy2.robjects import r

    CL = importr("ChainLadder")
except:
    pass


def mack_r(data, alpha, est_sigma):
    return r(
        'mack<-MackChainLadder({},alpha={}, est.sigma="{}")'.format(
            data, alpha, est_sigma
        )
    )


def mack_p(data, average, est_sigma):
    return cl.Development(average=average, sigma_interpolation=est_sigma).fit_transform(
        cl.load_sample(data)
    )


data = ["RAA", "GenIns", "MW2014"]
averages = [("simple", 0), ("volume", 1), ("regression", 2)]
est_sigma = [("mack", "Mack"), ("log-linear", "log-linear")]




@pytest.mark.r
@pytest.mark.parametrize("data", data)
@pytest.mark.parametrize("averages", averages)
@pytest.mark.parametrize("est_sigma", est_sigma)
def test_mack_ldf(data, averages, est_sigma, atol):
    p = mack_p(data, averages[0], est_sigma[0]).ldf_
    xp = p.get_array_module()
    r = xp.array(mack_r(data, averages[1], est_sigma[1]).rx("f"))[:, :-1]
    assert xp.allclose(r, p.values[0, 0, :, :], atol=atol)


@pytest.mark.r
@pytest.mark.parametrize("data", data)
@pytest.mark.parametrize("averages", averages)
@pytest.mark.parametrize("est_sigma", est_sigma)
def test_mack_sigma(data, averages, est_sigma, atol):
    p = mack_p(data, averages[0], est_sigma[0]).sigma_.set_backend(
        "numpy", inplace=True
    )
    xp = p.get_array_module()
    r = xp.array(mack_r(data, averages[1], est_sigma[1]).rx("sigma"))
    assert xp.allclose(r, p.values[0, 0, :, :], atol=atol)


@pytest.mark.r
@pytest.mark.parametrize("data", data)
@pytest.mark.parametrize("averages", averages)
@pytest.mark.parametrize("est_sigma", est_sigma)
def test_mack_std_err(data, averages, est_sigma, atol):
    p = mack_p(data, averages[0], est_sigma[0]).std_err_
    xp = p.get_array_module()
    r = xp.array(mack_r(data, averages[1], est_sigma[1]).rx("f.se"))
    assert xp.allclose(r, p.values[0, 0, :, :], atol=atol)



import chainladder as cl
import numpy as np
import pandas as pd


def test_constant_cdf(raa):
    dev = cl.Development().fit(raa)
    xp = dev.ldf_.get_array_module()
    link_ratios = {
        (num + 1) * 12: item for num, item in enumerate(dev.ldf_.values[0, 0, 0, :])
    }
    dev_c = cl.DevelopmentConstant(patterns=link_ratios, style="ldf").fit(raa)
    assert xp.allclose(dev.cdf_.values, dev_c.cdf_.values, atol=1e-5)


def test_constant_ldf(raa):
    dev = cl.Development().fit(raa)
    xp = dev.ldf_.get_array_module()
    link_ratios = {
        (num + 1) * 12: item for num, item in enumerate(dev.ldf_.values[0, 0, 0, :])
    }
    dev_c = cl.DevelopmentConstant(patterns=link_ratios, style="ldf").fit(raa)
    assert xp.allclose(dev.ldf_.values, dev_c.ldf_.values, atol=1e-5)

def test_constant_callable(clrd, atol):
    agway = clrd.loc['Agway Ins Co', 'CumPaidLoss']
    def paid_cdfs(x):
        """ A function that returns different CDFs depending on a specified LOB """
        cdfs = {
        'comauto': [3.832, 1.874, 1.386, 1.181, 1.085, 1.043, 1.022, 1.013, 1.007, 1],
        'medmal': [24.168, 4.127, 2.103, 1.528, 1.275, 1.161, 1.088, 1.047, 1.018, 1],
        'othliab': [10.887, 3.416, 1.957, 1.433, 1.231, 1.119, 1.06, 1.031, 1.011, 1],
        'ppauto': [2.559, 1.417, 1.181, 1.084, 1.04, 1.019, 1.009, 1.004, 1.001, 1],
        'prodliab': [13.703, 5.613, 2.92, 1.765, 1.385, 1.177, 1.072, 1.034, 1.008, 1],
        'wkcomp': [4.106, 1.865, 1.418, 1.234, 1.141, 1.09, 1.056, 1.03, 1.01, 1]}
        patterns = pd.DataFrame(cdfs, index=range(12, 132, 12)).T
        return patterns.loc[x.loc['LOB']].to_dict()
    model = cl.DevelopmentConstant(patterns=paid_cdfs, callable_axis=1, style='cdf')
    assert abs(model.fit_transform(agway).cdf_.loc['comauto'].iloc[..., 0].sum() - 3.832) < atol



import numpy as np
import chainladder as cl
import pytest


@pytest.fixture
def triangle_data():
    clrd = cl.load_sample("clrd")[["CumPaidLoss", "EarnedPremDIR"]]
    clrd = clrd[clrd["LOB"] == "wkcomp"]
    return clrd


@pytest.fixture
def estimators():
    bcl = cl.Chainladder()
    bf = cl.BornhuetterFerguson()
    cc = cl.CapeCod()

    estimators = [("bcl", bcl), ("bf", bf), ("cc", cc)]

    return estimators


array_weight = np.array([[1, 2, 3]] * 4 + [[0, 0.5, 0.5]] * 3 + [[0, 0, 1]] * 3)

list_weight = [[[1, 2, 3]] * 4 + [[0, 0.5, 0.5]] * 3 + [[0, 0, 1]] * 3]

callable_weight = lambda origin: np.where(
    origin.year < 1992,
    (1, 2, 3),
    np.where(origin.year > 1994, (0, 0, 1), (0, 0.5, 0.5)),
)

dict_weight = {
    "1992": (0, 0.5, 0.5),
    "1993": (0, 0.5, 0.5),
    "1994": (0, 0.5, 0.5),
    "1995": (0, 0, 1),
    "1996": (0, 0, 1),
    "1997": (0, 0, 1),
}


@pytest.fixture(params=[array_weight, list_weight, callable_weight, dict_weight])
def weights(request):
    return request.param


def test_voting_ultimate(triangle_data, estimators, weights):
    bcl_ult = (
        cl.Chainladder()
        .fit(
            triangle_data["CumPaidLoss"].sum(),
        )
        .ultimate_
    )
    bf_ult = (
        cl.BornhuetterFerguson()
        .fit(
            triangle_data["CumPaidLoss"].sum(),
            sample_weight=triangle_data["EarnedPremDIR"].sum().latest_diagonal,
        )
        .ultimate_
    )
    cc_ult = (
        cl.CapeCod()
        .fit(
            triangle_data["CumPaidLoss"].sum(),
            sample_weight=triangle_data["EarnedPremDIR"].sum().latest_diagonal,
        )
        .ultimate_
    )

    vot_ult = (
        cl.VotingChainladder(
            estimators=estimators, weights=weights, default_weighting=(1, 2, 3)
        )
        .fit(
            triangle_data["CumPaidLoss"].sum(),
            sample_weight=triangle_data["EarnedPremDIR"].sum().latest_diagonal,
        )
        .ultimate_
    )

    direct_weight = np.array([[1, 2, 3]] * 4 + [[0, 0.5, 0.5]] * 3 + [[0, 0, 1]] * 3)
    direct_weight = direct_weight[..., np.newaxis]

    assert (
        abs(
            (
                (
                    bcl_ult * direct_weight[..., 0, :]
                    + bf_ult * direct_weight[..., 1, :]
                    + cc_ult * direct_weight[..., 2, :]
                )
                / direct_weight.sum(axis=-2)
            ).sum()
            - vot_ult.sum()
        )
        < 1
    )


def test_different_backends(triangle_data, estimators, weights):
    model = cl.VotingChainladder(
        estimators=estimators, weights=weights, default_weighting=(1, 2, 3)
    ).fit(
        triangle_data["CumPaidLoss"].sum().set_backend("numpy"),
        sample_weight=triangle_data["EarnedPremDIR"]
        .sum()
        .latest_diagonal.set_backend("numpy"),
    )
    assert (
        abs(
            (
                model.predict(
                    triangle_data["CumPaidLoss"].sum().set_backend("sparse"),
                    sample_weight=triangle_data["EarnedPremDIR"]
                    .sum()
                    .latest_diagonal.set_backend("sparse"),
                ).ultimate_.sum()
                - model.ultimate_.sum()
            )
        )
        < 1
    )


def test_weight_broadcasting(triangle_data, estimators, weights):
    mid_dim_weights = np.array(
        [[[1, 2, 3]] * 4 + [[0, 0.5, 0.5]] * 3 + [[0, 0, 1]] * 3] * 1
    )
    max_dim_weights = np.array(mid_dim_weights * 132)

    min_dim_ult = (
        cl.VotingChainladder(estimators=estimators, weights=weights)
        .fit(
            triangle_data["CumPaidLoss"],
            sample_weight=triangle_data["EarnedPremDIR"].latest_diagonal,
        )
        .ultimate_.sum()
    )
    mid_dim_ult = (
        cl.VotingChainladder(estimators=estimators, weights=mid_dim_weights)
        .fit(
            triangle_data["CumPaidLoss"],
            sample_weight=triangle_data["EarnedPremDIR"].latest_diagonal,
        )
        .ultimate_.sum()
    )
    max_dim_ult = (
        cl.VotingChainladder(estimators=estimators, weights=max_dim_weights)
        .fit(
            triangle_data["CumPaidLoss"],
            sample_weight=triangle_data["EarnedPremDIR"].latest_diagonal,
        )
        .ultimate_.sum()
    )
    assert abs(min_dim_ult - mid_dim_ult - max_dim_ult) < 1


def test_voting(raa):
    cl_ult = cl.Chainladder().fit(raa).ultimate_
    apriori = cl_ult * 0 + (float(cl_ult.sum()) / 10)
    bcl = cl.Chainladder()
    bf = cl.BornhuetterFerguson()
    cc = cl.CapeCod()
    estimators = [("bcl", bcl), ("bf", bf), ("cc", cc)]
    weights = np.array([[0.6, 0.2, 0.2]] * 4 + [[0, 0.5, 0.5]] * 3 + [[0, 0, 1]] * 3)
    vot = cl.VotingChainladder(estimators=estimators, weights=weights)
    vot.fit(raa, sample_weight=apriori)
    assert (
        vot.ultimate_.to_frame().round(2).values.flatten()
        == [
            18834.00,
            16875.50,
            24058.53,
            28542.58,
            28236.84,
            19905.32,
            18947.25,
            23106.94,
            20004.50,
            21605.83,
        ]
    ).all()



import chainladder as cl

def test_basic_odp_cl(genins):
    assert abs(
        (cl.Chainladder().fit(genins).ultimate_ -
         cl.Chainladder().fit(cl.TweedieGLM().fit_transform(genins)).ultimate_) /
        genins.latest_diagonal).max()< 1e-2



import chainladder as cl


def test_mcl_ult():
    mcl = cl.load_sample("mcl")
    dev = cl.Development().fit_transform(mcl)
    cl_traditional = cl.Chainladder().fit(dev).ultimate_
    dev_munich = cl.MunichAdjustment(
        paid_to_incurred=[("paid", "incurred")]
    ).fit_transform(dev)
    cl_munich = cl.Chainladder().fit(dev_munich).ultimate_


def test_mcl_rollforward():
    mcl = cl.load_sample("mcl")
    mcl_prior = mcl[mcl.valuation < mcl.valuation_date]
    munich = cl.MunichAdjustment(paid_to_incurred=[("paid", "incurred")]).fit(mcl_prior)
    new = munich.transform(mcl)
    cl.Chainladder().fit(new).ultimate_



import chainladder as cl
import pytest

raa = cl.load_sample("RAA")

def test_val_corr_total_true():
    assert raa.valuation_correlation(p_critical=0.5, total=True)

def test_val_corr_total_false():
    assert raa.valuation_correlation(p_critical=0.5, total=False)

def test_dev_corr():
    assert raa.development_correlation(p_critical=0.5)

def test_dev_corr_sparse():
    assert raa.set_backend('sparse').development_correlation(p_critical=0.5)

def test_validate_critical():
    with pytest.raises(ValueError):
        raa.valuation_correlation(p_critical=1.5, total=True)


import chainladder as cl
import numpy as np


def test_schmidt():
    tri = cl.load_sample("ia_sample")
    xp = np if tri.array_backend == "sparse" else tri.get_array_module()
    ia = cl.IncrementalAdditive()
    ia_transform = ia.fit_transform(
        tri.iloc[0, 0], sample_weight=tri.iloc[0, 1].latest_diagonal
    )
    answer = ia_transform.incremental_.incr_to_cum().values[0, 0, :, -1]
    answer_zeta = ia_transform.zeta_.values[0, 0, 0, :]
    answer_cum_zeta = ia_transform.cum_zeta_.values[0, 0, 0, :]
    check = xp.array(
        [
            3483.0,
            4007.84795031,
            4654.36196862,
            5492.00685523,
            6198.10197128,
            7152.82539296,
        ]
    )
    check_zeta = xp.array(
        [0.24321225, 0.22196026, 0.15397836, 0.14185271, 0.09067327, 0.03677019]
    )
    check_cum_zeta = xp.array(
        [0.88844704, 0.64523479, 0.42327453, 0.26929617, 0.12744346, 0.03677019]
    )
    assert (
        xp.allclose(answer, check, atol=1e-5)
        & xp.allclose(answer_zeta, check_zeta, atol=1e-8)
        & xp.allclose(answer_cum_zeta, check_cum_zeta, atol=1e-8)
    )


def test_IBNR_methods():
    tri = cl.load_sample("ia_sample")
    incr_est = cl.IncrementalAdditive().fit(
        tri["loss"], sample_weight=tri["exposure"].latest_diagonal
    )
    incr_trans = incr_est.transform(tri["loss"])
    incr_tri = incr_est.incremental_.incr_to_cum()
    incr_ult = incr_tri[incr_tri.development == incr_tri.development.max()]
    cl_est = cl.Chainladder().fit(incr_trans)
    bf_est = cl.BornhuetterFerguson(apriori=1).fit(incr_trans, sample_weight=incr_ult)
    incr_result = np.round(incr_est.incremental_.values, 5)
    cl_result = np.round(cl_est.full_triangle_.cum_to_incr().values[..., :-2], 5)
    bf_result = np.round(bf_est.full_triangle_.cum_to_incr().values[..., :-2], 5)
    assert np.all(incr_result == cl_result) & np.all(incr_result == bf_result)


def test_pipeline():
    clrd = cl.load_sample("clrd").groupby("LOB")[["IncurLoss", "CumPaidLoss"]].sum()
    dev = cl.Development().fit_transform(clrd)
    ult = cl.Chainladder().fit(clrd)
    dev1 = cl.IncrementalAdditive(
        n_periods=7,
        drop_valuation=1995,
        drop=("1992", 12),
        drop_above=1.05,
        drop_below=-1,
        drop_high=1,
        drop_low=1,
    ).fit(clrd, sample_weight=ult.ultimate_ * 3)
    pipe = cl.Pipeline(
        steps=[
            ("n_periods", cl.IncrementalAdditive(n_periods=7)),
            ("drop_valuation", cl.IncrementalAdditive(drop_valuation=1995)),
            ("drop", cl.IncrementalAdditive(drop=("1992", 12))),
            (
                "drop_abovebelow",
                cl.IncrementalAdditive(drop_above=1.05, drop_below=0.95),
            ),
            ("drop_hilo", cl.IncrementalAdditive(drop_high=1, drop_low=1)),
        ]
    )
    dev2 = pipe.fit(X=clrd, sample_weight=ult.ultimate_ * 3)
    assert np.array_equal(
        dev1.zeta_.values, dev2.named_steps.drop_hilo.zeta_.values, True
    )



import chainladder as cl
import numpy as np


def test_basic_case_outstanding():
    tri = cl.load_sample("usauto")
    m = cl.CaseOutstanding(paid_to_incurred=("paid", "incurred")).fit(tri)
    out = cl.Chainladder().fit(m.fit_transform(tri))
    a = (out.full_triangle_["incurred"] - out.full_triangle_["paid"]).iloc[
        ..., -1, :9
    ] * m.paid_ldf_.values
    b = (out.full_triangle_["paid"].cum_to_incr().iloc[..., -1, 1:10]).values
    assert (a - b).max() < 1e-6


def test_outstanding_friedland_example():
    usauto = cl.load_sample("usauto")
    model = cl.CaseOutstanding(
        paid_to_incurred=("paid", "incurred"), paid_n_periods=3, case_n_periods=3
    ).fit(usauto)

    expected_paid_ldf = np.array(
        [
            [
                0.833,
                0.701,
                0.714,
                0.714,
                0.653,
                0.631,
                0.553,
                0.437,
                0.524,
            ]
        ]
    )
    assert (
        model.paid_ldf_.to_frame(origin_as_datetime=False).values - expected_paid_ldf
        < 0.001
    ).all()

    expected_case_ldf = np.array(
        [
            [
                0.526,
                0.566,
                0.528,
                0.486,
                0.511,
                0.555,
                0.652,
                0.674,
                0.580,
            ]
        ]
    )
    assert (
        model.case_ldf_.to_frame(origin_as_datetime=False).values - expected_case_ldf
        < 0.001
    ).all()



import numpy as np
import chainladder as cl
import pytest


def test_full_slice():
    assert (
        cl.Development().fit_transform(cl.load_sample("GenIns")).ldf_
        == cl.Development(n_periods=1000).fit_transform(cl.load_sample("GenIns")).ldf_
    )


def test_full_slice2():
    assert (
        cl.Development().fit_transform(cl.load_sample("GenIns")).ldf_
        == cl.Development(n_periods=[1000] * (cl.load_sample("GenIns").shape[3] - 1))
        .fit_transform(cl.load_sample("GenIns"))
        .ldf_
    )


def test_drop1(raa):
    assert (
        cl.Development(drop=("1982", 12)).fit(raa).ldf_.values[0, 0, 0, 0]
        == cl.Development(drop_high=[True] + [False] * 8)
        .fit(raa)
        .ldf_.values[0, 0, 0, 0]
    )


def test_drop2(raa):
    assert (
        cl.Development(drop_valuation="1981").fit(raa).ldf_.values[0, 0, 0, 0]
        == cl.Development(drop_low=[True] + [False] * 8)
        .fit(raa)
        .ldf_.values[0, 0, 0, 0]
    )


def test_n_periods():
    d = cl.load_sample("usauto")["incurred"]
    xp = np if d.array_backend == "sparse" else d.get_array_module()
    return xp.all(
        xp.around(
            xp.unique(
                cl.Development(n_periods=3, average="volume").fit(d).ldf_.values,
                axis=-2,
            ),
            3,
        ).flatten()
        == xp.array([1.164, 1.056, 1.027, 1.012, 1.005, 1.003, 1.002, 1.001, 1.0])
    )


def test_drophighlow():
    raa = cl.load_sample("raa")

    lhs = np.round(cl.Development(drop_high=0).fit(raa).cdf_.values, 4).flatten()
    rhs = np.array(
        [8.9202, 2.974, 1.8318, 1.4414, 1.2302, 1.1049, 1.0604, 1.0263, 1.0092]
    )
    assert np.all(lhs == rhs)

    lhs = np.round(
        cl.Development(drop_high=[True, False, True, False]).fit(raa).cdf_.values, 4
    ).flatten()
    rhs = np.array(
        [8.0595, 2.8613, 1.7624, 1.4414, 1.2302, 1.1049, 1.0604, 1.0263, 1.0092]
    )
    assert np.all(lhs == rhs)

    lhs = np.round(
        cl.Development(drop_high=[2, 3, 3, 3], drop_low=[0, 1, 0], preserve=2)
        .fit(raa)
        .cdf_.values,
        4,
    ).flatten()
    rhs = np.array(
        [
            5.7403,
            2.2941,
            1.5617,
            1.3924,
            1.2302,
            1.1049,
            1.0604,
            1.0263,
            1.0092,
        ]
    )
    assert np.all(lhs == rhs)

    lhs = np.round(cl.Development(drop_high=1).fit(raa).cdf_.values, 4).flatten()
    rhs = np.array(
        [7.2190, 2.5629, 1.6592, 1.3570, 1.1734, 1.0669, 1.0419, 1.0121, 1.0092]
    )
    assert np.all(lhs == rhs)

    lhs = np.round(
        cl.Development(drop_high=1, drop_low=1).fit(raa).cdf_.values, 4
    ).flatten()
    rhs = np.array(
        [9.0982, 2.8731, 1.8320, 1.4713, 1.2522, 1.0963, 1.0604, 1.0263, 1.0092]
    )
    assert np.all(lhs == rhs)

    lhs = np.round(
        cl.Development(drop_high=[2, 1, 1], drop_low=1).fit(raa).cdf_.values, 4
    ).flatten()
    rhs = np.array(
        [8.4905, 3.0589, 1.9504, 1.5664, 1.3142, 1.1403, 1.0822, 1.0426, 1.0092]
    )
    assert np.all(lhs == rhs)

    lhs = np.round(
        cl.Development(drop_high=1, drop_low=1, n_periods=5).fit(raa).cdf_.values, 4
    ).flatten()
    rhs = np.array(
        [16.3338, 3.2092, 1.8124, 1.4793, 1.2522, 1.0963, 1.0604, 1.0263, 1.0092]
    )
    assert np.all(lhs == rhs)

    tri = cl.load_sample("prism")["Paid"].sum().grain("OYDQ")
    no_drop = cl.Development().fit_transform(tri).cdf_.to_frame().values
    drop_high = cl.Development(drop_high=True).fit_transform(tri).cdf_.to_frame().values
    drop_low = cl.Development(drop_low=True).fit_transform(tri).cdf_.to_frame().values
    assert (drop_low >= no_drop).all()
    assert (no_drop >= drop_high).all()


def test_dropabovebelow():
    raa = cl.load_sample("raa")

    lhs = np.round(cl.Development(drop_above=40.0).fit(raa).cdf_.values, 4).flatten()
    rhs = np.array(
        [8.3771, 2.9740, 1.8318, 1.4414, 1.2302, 1.1049, 1.0604, 1.0263, 1.0092]
    )
    assert np.all(lhs == rhs)

    lhs = np.round(cl.Development(drop_above=1.2).fit(raa).cdf_.values, 4).flatten()
    rhs = np.array(
        [7.6859, 2.5625, 1.5784, 1.4072, 1.2302, 1.1049, 1.0604, 1.0263, 1.0092]
    )
    assert np.all(lhs == rhs)

    lhs = np.round(
        cl.Development(drop_above=1.2, drop_below=1.05).fit(raa).cdf_.values, 4
    ).flatten()
    rhs = np.array(
        [8.4983, 2.8334, 1.7452, 1.5560, 1.3602, 1.1802, 1.0604, 1.0263, 1.0092]
    )
    assert np.all(lhs == rhs)

    lhs = np.round(
        cl.Development(drop_above=[40.0], drop_below=[0.0, 0.0, 1.05, 1.7])
        .fit(raa)
        .cdf_.values,
        4,
    ).flatten()
    rhs = np.array(
        [8.3771, 2.9740, 1.8318, 1.4414, 1.2302, 1.1049, 1.0604, 1.0263, 1.0092]
    )
    assert np.all(lhs == rhs)

    lhs = np.round(
        cl.Development(drop_above=[40.0], drop_below=[0.0, 0.0, 1.05, 1.2])
        .fit(raa)
        .cdf_.values,
        4,
    ).flatten()
    rhs = np.array(
        [8.9773, 3.1871, 1.9631, 1.5447, 1.2302, 1.1049, 1.0604, 1.0263, 1.0092]
    )
    assert np.all(lhs == rhs)


def test_drop_valuation():
    raa = cl.load_sample("raa")
    assert (
        cl.Development(drop_valuation="1981-12-31").fit_transform(raa).cdf_
        != cl.Development(drop_valuation="1982-12-31").fit_transform(raa).cdf_
    )
    assert (
        cl.Development(drop_valuation="1982-12-31").fit_transform(raa).cdf_
        != cl.Development(drop_valuation="1983-12-31").fit_transform(raa).cdf_
    )
    assert (
        cl.Development(drop_valuation="1981-12-31").fit_transform(raa).cdf_
        != cl.Development(drop_valuation="1983-12-31").fit_transform(raa).cdf_
    )

    quarterly = cl.load_sample("quarterly")
    assert (
        cl.Development().fit_transform(quarterly["incurred"]).cdf_
        != cl.Development(drop_valuation="1995-03-31")
        .fit_transform(quarterly["incurred"])
        .cdf_
    )
    assert (
        cl.Development().fit_transform(quarterly["incurred"]).cdf_
        != cl.Development(drop_valuation="1995-06-30")
        .fit_transform(quarterly["incurred"])
        .cdf_
    )
    assert (
        cl.Development(drop_valuation="1995-03-31")
        .fit_transform(quarterly["incurred"])
        .cdf_
        != cl.Development(drop_valuation="1995-06-30")
        .fit_transform(quarterly["incurred"])
        .cdf_
    )
    assert (
        cl.Development(drop_valuation="1995-06-30")
        .fit_transform(quarterly["incurred"])
        .cdf_
        != cl.Development(drop_valuation="1995-09-30")
        .fit_transform(quarterly["incurred"])
        .cdf_
    )


def test_assymetric_development(atol):
    quarterly = cl.load_sample("quarterly")["paid"]
    xp = np if quarterly.array_backend == "sparse" else quarterly.get_array_module()
    dev = cl.Development(n_periods=1, average="simple").fit(quarterly)
    dev2 = cl.Development(n_periods=1, average="regression").fit(quarterly)
    assert xp.allclose(dev.ldf_.values, dev2.ldf_.values, atol=atol)


def test_hilo_multiple_indices(clrd):
    tri = clrd.groupby("LOB")["CumPaidLoss"].sum()
    assert (
        cl.Development(n_periods=5).fit(tri).ldf_.loc["wkcomp"]
        == cl.Development(n_periods=5).fit(tri.loc["wkcomp"]).ldf_
    )
    assert (
        cl.Development(drop_low=2).fit(tri).ldf_.loc["wkcomp"]
        == cl.Development(drop_low=2).fit(tri.loc["wkcomp"]).ldf_
    )


def test_new_drop_1(clrd):
    clrd = clrd.groupby("LOB")[["IncurLoss", "CumPaidLoss"]].sum()
    # n_periods
    compare_new_drop(cl.Development(n_periods=4).fit(clrd), clrd)


def test_new_drop_2(clrd):
    clrd = clrd.groupby("LOB")[["IncurLoss", "CumPaidLoss"]].sum()
    # single drop and drop_valuation
    compare_new_drop(
        cl.Development(drop=("1992", 12), drop_valuation=1993).fit(clrd), clrd
    )


def test_new_drop_3(clrd):
    clrd = clrd.groupby("LOB")[["IncurLoss", "CumPaidLoss"]].sum()
    # multiple drop and drop_valuation
    compare_new_drop(
        cl.Development(
            drop=[("1992", 12), ("1996", 24)], drop_valuation=[1993, 1995]
        ).fit(clrd),
        clrd,
    )


def test_new_drop_4(clrd):
    clrd = clrd.groupby("LOB")[["IncurLoss", "CumPaidLoss"]].sum()
    # drop_hi/low without preserve
    compare_new_drop(cl.Development(drop_high=1, drop_low=1).fit(clrd), clrd)


def test_new_drop_5(clrd):
    clrd = clrd.groupby("LOB")[["IncurLoss", "CumPaidLoss"]].sum()
    # drop_hi/low without preserve
    compare_new_drop(
        cl.Development(drop_high=1, drop_low=1, preserve=3).fit(clrd), clrd
    )


def test_new_drop_5a(clrd):
    clrd = clrd.groupby("LOB")[["IncurLoss", "CumPaidLoss"]].sum()
    # drop_hi/low without preserve
    assert np.array_equal(
        cl.Development(drop_high=1, drop_low=1, preserve=3)
        ._set_weight_func(clrd.age_to_age, clrd.age_to_age)
        .values,
        cl.Development(
            drop_high=True,
            drop_low=[True, True, True, True, True, True, True, True, True],
            preserve=3,
        )
        ._set_weight_func(clrd.age_to_age)
        .values,
        True,
    )


def test_new_drop_6(clrd):
    clrd = clrd.groupby("LOB")[["IncurLoss", "CumPaidLoss"]].sum()
    # drop_above/below without preserve
    compare_new_drop(cl.Development(drop_above=1.01, drop_below=0.95).fit(clrd), clrd)


def test_new_drop_7(clrd):
    clrd = clrd.groupby("LOB")[["IncurLoss", "CumPaidLoss"]].sum()
    # drop_above/below with preserve
    compare_new_drop(
        cl.Development(drop_above=1.01, drop_below=0.95, preserve=3).fit(clrd), clrd
    )


def test_new_drop_8():
    tri = cl.load_sample("prism")["Paid"].sum().grain("OYDQ")

    try:
        cl.Development(drop_high=False).fit_transform(tri)
    except:
        assert False


def test_new_drop_9():
    tri = cl.load_sample("prism")["Paid"].sum().grain("OYDQ")

    lhs = cl.Development(drop_high=True).fit(tri).cdf_.to_frame().fillna(0).values
    rhs = cl.Development(drop_high=1).fit(tri).cdf_.to_frame().fillna(0).values
    assert (lhs == rhs).all()


@pytest.mark.xfail
def test_new_drop_10():
    data = {
        "valuation": [
            1981,
            1982,
            1983,
            1984,
            1985,
            1982,
            1983,
            1984,
            1985,
        ],
        "origin": [
            1981,
            1982,
            1983,
            1984,
            1985,
            1981,
            1982,
            1983,
            1984,
        ],
        "values": [
            100,
            200,
            300,
            400,
            500,
            200,
            200,
            300,
            800,
        ],
    }

    tri = cl.Triangle(
        pd.DataFrame(data),
        origin="origin",
        development="valuation",
        columns=["values"],
        cumulative=True,
    )

    assert np.round(
        cl.Development(drop_high=1).fit(tri).cdf_.to_frame().values.flatten()[0], 4
    ) == (200 + 300 + 800) / (200 + 300 + 400)

    assert (
        np.round(
            cl.Development(drop_high=2).fit(tri).cdf_.to_frame().values.flatten()[0], 4
        )
        == 1.0000
    )


def compare_new_drop(dev, tri):
    assert np.array_equal(
        dev._set_weight_func(tri.age_to_age, tri.age_to_age).values,
        dev.transform(tri).age_to_age.values * 0 + 1,
        True,
    )


def test_4d_drop(clrd):
    clrd = clrd.groupby("LOB").sum()[["CumPaidLoss", "IncurLoss"]]
    assert (
        cl.Development(n_periods=4).fit_transform(clrd.iloc[0, 0]).link_ratio
        == cl.Development(n_periods=4).fit_transform(clrd).link_ratio.iloc[0, 0]
    )


def test_pipeline(clrd):
    clrd = clrd.groupby("LOB")[["IncurLoss", "CumPaidLoss"]].sum()
    dev1 = cl.Development(
        n_periods=7,
        drop_valuation=1995,
        drop=("1992", 12),
        drop_above=1.05,
        drop_below=0.95,
        drop_high=1,
        drop_low=1,
    ).fit(clrd)
    pipe = cl.Pipeline(
        steps=[
            ("n_periods", cl.Development(n_periods=7)),
            ("drop_valuation", cl.Development(drop_valuation=1995)),
            ("drop", cl.Development(drop=("1992", 12))),
            ("drop_abovebelow", cl.Development(drop_above=1.05, drop_below=0.95)),
            ("drop_hilo", cl.Development(drop_high=1, drop_low=1)),
        ]
    )
    dev2 = pipe.fit(X=clrd)
    assert np.array_equal(
        dev1.w_v2_.values, dev2.named_steps.drop_hilo.w_v2_.values, True
    )
